#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import os
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

DEBUG = False


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

def flatten(list):

    '''
    Flatten a multi-dimensional list into a single list.

    :param list(list(any)) list: the multi-dimensional list to flatten.
    :return list(any) newList: the flattened, single list.
    '''

    newList = []
    for i in list:
        for j in i:
            newList.append(j)
    return newList


#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from .spriteTextureList import SpriteTextureList

class SpriteImage():

    '''
    Maps string states to a pygamepal.SpriteTextureList object containing textures and associated information.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/spriteImageExample.gif?raw=true

    `Example SpriteImage code`_.

    .. _Example SpriteImage code: https://github.com/rik-cross/pygamepal/blob/main/examples/spriteImageExample.py

    Parameters for the first provided state:

    :param pygame.texture textures: 0 or more textures to add to the first SpriteTextureList. Note that if more than 1 texture is added, the spriteImage becomes an animation.
    :param str state: The state name to associate with added textures (default = None).
    :param int animationDelay: The amount of time to display each texture (default = 8).
    :param bool loop: Loop animation (ignored if only one image provided for a state, default = True).
    :param bool hFlip: Horizontally flip images (default = False).
    :param bool vFlip: Vertically flip textures (default = False).
    :param (int, int) offset: Used to ignore part of the image when specifying a drawing position. Useful for sprites with whitespace or characters holding tools, etc. (default = (0, 0)).
    
    Parameters for all states:
    
    :param bool visible: Show image (default = True).
    :param int alpha: The transparency value (between 0 and 255, default = 255).
    :param bool pause: Stops animation (ignored for states with only one image, default = False).
    '''

    def __init__(self,
        *textures,
        state = None,
        animationDelay = 8,
        loop = True,
        hFlip = False,
        vFlip = False,
        offset = (0, 0),
        visible = True,
        alpha = 255,
        pause = False
    ):

        # reset the object
        self.reset()
        # set overall visibility and alpha
        self.visible = visible
        self.alpha = alpha
        self.pause = pause
        # associate textures with a state, along with other info 
        if len(textures) > 0:
            self.addTextures(*textures, state = state, animationDelay = animationDelay, loop = loop, hFlip = hFlip, vFlip = vFlip, offset = offset)

    def update(self, deltaTime = 1):

        '''
        Updates the current sprite animation. Must be called once per frame.

        :param float deltaTime: Time elapsed since last call (default = 1).
        '''

        # only animate if there's at least one image and it's not paused
        if self._currentState is None or self.pause:
            return
        
        # increment timer
        self._animationTimer += 1

        # advance animation if timer reaches the delay value
        if self._animationTimer >= self._textureLists[self._currentState]._animationDelay:
            self._animationTimer = 0
            self._animationIndex += 1
            # loop back to the first texture if specified
            if self._animationIndex > len(self._textureLists[self._currentState]._textures) - 1:
                if self._textureLists[self._currentState]._loop:
                    self._animationIndex = 0
                else:
                    self._animationIndex = len(self._textureLists[self._currentState]._textures) - 1

    def draw(self, surface, position):

        '''
        Draws the current frame of the current state's animation. Must be called once per frame.
        
        :param pygame.surface surface: The surface to draw to.
        :param (int, int) position: The (x, y) position to draw to (default = (0, 0)).
        '''

        # don't draw if there's nothing to draw
        if self._currentState is None or self._textureLists[self._currentState] is None or self.visible is False:
            return
        # get the current animation frame
        currentTexture = self._textureLists[self._currentState]._textures[self._animationIndex]
        # set the texture alpha
        currentTexture.set_alpha(self.alpha)
        # draw (optionally flipped) texture
        surface.blit(pygame.transform.flip(currentTexture,
                    self._textureLists[self._currentState]._hFlip,
                    self._textureLists[self._currentState]._vFlip),
                    (position[0] - self._textureLists[self._currentState]._offset[0],
                     position[1] - self._textureLists[self._currentState]._offset[1],
                     currentTexture.get_width(),
                     currentTexture.get_height()))

    def addTextures(self, *textures, state = None, animationDelay = 8, loop = True, hFlip = False, vFlip = False, offset = (0, 0)):

        '''
        Add one or more textures to a (new or existing) state.

        :param pygame.texture textures: 0 or more textures to add to the first SpriteTextureList. Note that if more than 1 texture is added, the spriteImage becomes an animation.
        :param str state: The state name to associate with added textures (default = None).
        :param int animationDelay: The amount of time to display each texture (default = 8).
        :param bool loop: Loop animation (ignored if only one image provided for a state, default = True).
        :param bool hFlip: Horizontally flip images (default = False).
        :param bool vFlip: Vertically flip textures (default = False).
        :param (int, int) offset: Used to ignore part of the image when specifying a drawing position. Useful for sprites with whitespace or characters holding tools, etc. (default = (0, 0)).
        '''

        # allow textures with no attached state (for single-state images/animations)
        if state is None:
            state = 'default'
        
        # get existing textureList or create a new one
        if state not in self._textureLists:
            textureList = SpriteTextureList()
        else:
            textureList = self._sprites[state] 

        # add textures to list
        for texture in textures:
            texture = texture.convert_alpha()
            textureList._textures.append(texture)
        
        # add attributes for the current state
        textureList._animationDelay = animationDelay
        textureList._loop = loop
        textureList._hFlip = hFlip
        textureList._vFlip = vFlip
        textureList._offset = offset

        # add the texture list to the SpriteImage, stored
        # against the appropriate state
        self._textureLists[state] = textureList

        # set the state if this is the only one
        if len(self._textureLists) == 1:
            self._currentState = state

    def reset(self):

        '''
        Clear object states and textures, and reset the object to default values.
        '''

        self.visible = True
        self.alpha = 255
        # maps a state to a list of textures
        self._textureLists = {}
        self._currentState = None
        # keeps track of current animation frame and progress
        self._animationIndex = 0
        self._animationTimer = 0
        self.pause = False

    def _resetState(self):

        '''
        Reset the current sprite animation.
        '''

        self._animationIndex = 0
        self._animationTimer = 0
    
    #
    # properties
    #

    # state

    @property
    def state(self):
        '''
        Get / set the current state.
        '''
        return self._currentState
    
    @state.setter
    def state(self, value):
        # only change state if the new state is different
        if value == self._currentState:
            return
        # set the new state
        self._currentState = value
        # reset index and timer for the new state
        self._resetState()

    # alpha transparency

    @property
    def alpha(self):
        '''
        Get / set the current alpha value.
        '''
        return self._alpha
    
    @alpha.setter
    def alpha(self, value):
        if value < 0 or value > 255:
            raise ValueError('alpha must be between 0 and 255')
        self._alpha = value
    
    # current texture center

    @property
    def center(self):
        '''
        Get the center positon for the current texture.
        '''
        if self._currentState is None:
            return (0, 0)
        # get the current animation frame
        currentTexture = self._textureLists[self._currentState]._textures[self._animationIndex]
        return (currentTexture.get_width() / 2, currentTexture.get_height() / 2)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

# create a default 'system' font
pygame.font.init()
sysFont = pygame.font.SysFont(None, 24)
smallFont = pygame.font.SysFont(None, 14)
regularFont = pygame.font.SysFont(None, 24)
largeFont = pygame.font.SysFont(None, 34)

def drawText(surface, text,
             position = [0, 0],
             font = None,
             antialias = True,
             color = 'white', backgroundColor = None,
             centerX = False, centerY = False):
    
    '''
    Draw text.

    :param pygame.Surface surface: The surface to draw to.
    :param str text: The text to draw.
    :param (int, int) position: The (x, y) position to draw on the specified surface (default = (0, 0)).
    :param pygame.Font font: The font to draw the text in (default = pygamepal.sysFont).
    :param bool antialias: Antialias text (default = True).
    :param pygame.Color color: Text color (default = 'white').
    :param pygame.Color backgroundColor: Background text color (default = None).
    :param bool centerX: Center horizontally (default = False).
    :param bool centerY: Center vertically (default = False).
    '''

    # use the default 'system' font if none specified
    if font is None:
        font = sysFont

    # ensure the color is a pygame color
    color = pygame.Color(color)

    # create text surface
    textSurface = font.render(text, antialias, color, backgroundColor)
    #print(color)
    if len(color) > 3:
        textSurface.set_alpha(color[3])

    # center
    if centerX == True:
        position = (position[0] - textSurface.get_rect().width // 2, position[1])
    if centerY == True:
        position = (position[0], position[1] - textSurface.get_rect().height // 2)

    # draw text to surface
    surface.blit(textSurface, position)

def splitText(text = None, width = None, font = None):

    if font is None:
        font = sysFont
    
    #print('>>>font')
    #print(font)

    returnList = []
    h = font.get_height()
    
    words = text.split(' ')
    #print(words)

    position = 0
    currentLine = ''

    fin = False
    while not fin:
        currentLine += words[position] + ' '
        position += 1
        #print(currentLine)
        if font.size(currentLine)[0] >= width:
            #print('ggg')
            #print(currentLine.split(' ')[:-2])
            currentLine = ' '.join(currentLine.split(' ')[:-2]) + ' '
            position -= 1
            returnList.append(currentLine)
            currentLine = ''
        if position >= len(words):
            fin = True
    if len(currentLine) > 0:
        returnList.append(currentLine)
    # remove trailing space
    returnList[-1] = returnList[-1][:-1]
    #print(returnList[-1][-1])
    #print('***')
    #print(returnList[-1])
    #print(returnList)
    return returnList



#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class SpriteTextureList():

    '''
    Used for storing a list of textures (and other info) for a SpriteImage state.
    (Users of the PygamePal library should not need to use this class directly. Instead, it is used by pygamepal.SpriteImage).
    '''

    def __init__(self):
        self._textures = []
        self._animationDelay = 12
        self._loop = True
        self._hFlip = False
        self._vFlip = False
        self._offset = (0, 0)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
import os

from .globals import ROOT_DIR

class Light:

    '''
    A light to add to the pygamepal.Lighting system.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/lightingExample.gif?raw=true
    
    `Example Lighting code`_.

    .. _Example Lighting code: https://github.com/rik-cross/pygamepal/blob/main/examples/lightingExample.py

    :param (int, int) position: Light position on the drawn surface.
    :param int radius: The radius of the light.
    :param str name: Light name (for getting and controlling if required, default = None).
    :param bool on: Light visibility (default = True).
    '''

    def __init__(self, position = (0, 0), radius = 100, name = None, on = True):
        self.position = list(position)
        self.name = name
        self.radius = radius
        self.on = True
    
    def toggle(self):

        '''
        Toggle the light visibility.
        '''
        
        self.on = not self.on
    
    #
    # properties
    #

    @property
    def radius(self):
        '''
        Get / set the radius value (is set to 0 if given negative value).
        '''
        return self._radius
    
    @radius.setter
    def radius(self, value):
        self._radius = max(0, value)

class Lighting:

    @property
    def lightLevel(self):
        '''
        Get / Set the ambient light level (between 0 (dark) and 1 (light), default = 0).
        '''
        return self._lightLevel
    
    @lightLevel.setter
    def lightLevel(self, value):
        self._lightLevel = min(max(0, value), 1)

    def __init__(self, surfaceSize, lightLevel = 0):
        
        # create a new lighting surface in the specified size
        self.surface = pygame.Surface(surfaceSize, pygame.SRCALPHA)
        self.lightLevel = lightLevel
        self.lightMask = pygame.image.load(os.path.join(ROOT_DIR, 'lightMask.png'))
        
        self.lights = []
        self.surface.fill( 'black' )

    def update(self, deltaTime = 1):

        '''
        Updates the lighting system, including all lights.

        :param float deltaTime: The elapsed time cince the last update (default = 1).
        '''
        
        pass

    def draw(self, surface):

        '''
        Draws all light in the lighting system.

        :param pygame.Surface surface: The surface to draw to.
        '''
        
        self.surface.fill('black')
        for l in [x for x in self.lights if x.on]:
            lightSize = (l.radius * 2, l.radius * 2)
            self.surface.blit(pygame.transform.scale(self.lightMask, lightSize), (l.position[0] - (lightSize[0] / 2), l.position[1] - (lightSize[1] / 2)), special_flags=pygame.BLEND_RGBA_SUB)
        surface.blit(self.surface, (0,0))
        self.surface.set_alpha(255 * (1 - self.lightLevel))

    def addLight(self, light):
        
        '''
        Adds a light to the lighting system.

        :param pygamepal.Light: The light to add.
        '''
        
        self.lights.append(light)

    def getLight(self, name):

        '''
        Get light by name. Returns None if no light exists.

        :param str name: The name of the light to get.
        '''
        
        for l in self.lights:
            if l.name == name:
                return l
        return None



#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

def easeLinear(x):

    '''
    Linear easing function.
    '''
    
    return x

def easeBounceOut(x):

    '''
    Ease bounce out easing function.
    '''
    
    x = x/100
    if x < 4 / 11:
        return (121 * x * x / 16) * 100
    elif x < 8 / 11:
        return ((363 / 40.0 * x * x) - (99 / 10.0 * x) + 17 / 5.0) * 100
    elif x < 9 / 10:
        return ((4356 / 361.0 * x * x) - (35442 / 1805.0 * x) + 16061 / 1805.0) * 100
    return ((54 / 5.0 * x * x) - (513 / 25.0 * x) + 268 / 25.0) * 100


#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from .drawText import drawText, splitText, sysFont
from .dialoguePage import DialoguePage

class Dialogue:

    '''
    Displays one or more pages of dialogue.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/dialogueExample.gif?raw=true

    `Example Dialogue code`_.

    .. _Example Dialogue code: https://github.com/rik-cross/pygamepal/blob/main/examples/dialogueExample.py

    :param (int, int) position: The (x, y) top-left position of the box (default = (0, 0)).
    :param (int, int) size: The (width, height) size of the box (default = (50, 50)).
    :param pygame.Font font: The text font (default = None, which uses pygamepal.sysFont).
    :param pygame.Color backgroundColor: Dialogue box background color (default = None).
    :param pygame.Color textColor: Text color (default = 'black').
    :param pygame.Color borderColor: Box and texture border color (default = 'black').
    :param int borderWidth: The width of the border, in pixels (default = 0).
    :param int paddingWidth: The space in pixels between elements (default = 10).
    :param str textEffect: Either 'none' or 'tick' (default = 'none').
    :param int tickSpeed: The delay between showing each character (if textEffect == 'tick' only, default = 4).
    :param str advanceTextSymbol: The string to show at the end of pages 0 to n-1 of dialogue (default = '>').
    :param pygame.Sound tickSound: Sound to play when advancing the text (if textEffect == 'tick' only, default = None).
    :param int borderRadius: Bevel on the edges of the borders (default = 0).
    :param bool visible: Vibisility of the dialogue box (default = True).
    :param function customDrawMethod: Provide an alternative draw() method (default = None).
    '''

    def __init__(self,
        position = (0, 0),
        size = (200, 100),
        font = None,
        backgroundColor = None,
        textColor = 'black',
        borderColor = 'black',
        borderWidth = 0,
        padding = 10,
        textEffect = 'none',
        tickSpeed = 4,
        advanceTextSymbol = '>',
        tickSound = None,
        borderRadius = 0,
        visible = True,
        custonDrawMethod = None
    ):

        self.position = position
        self.size = size
        
        if font is None:
            self.font = sysFont
        else:
            self.font = font

        self.backgroundColor = backgroundColor

        self.textColor = textColor
        self.borderColor = borderColor
        self.borderWidth = borderWidth
        self.padding = padding        
        self.textEffect = textEffect
        self.tickSpeed = tickSpeed
        self.advanceTextSymbol = advanceTextSymbol
        self.tickSound = tickSound
        self.borderRadius = borderRadius
        self.visible = visible
        self.customDrawMethod = custonDrawMethod

        self.pages = []
        self._pageIndex = 0
        self._characterIndex = 0
        self._tickTimer = 0
        self.complete = False

        # create a main dialogue box surface
        self._surface = pygame.Surface(self.size, pygame.SRCALPHA)

        # calculate the text position
        

    def update(self, deltaTime = 1):

        '''
        Updates the dialogue box, once per frame.

        :param float deltaTime: the game time elapsed since the last frame (default = 1).
        '''

        # only update if:
        # - there is at least one dialogue page
        # - the dialogue is not yet complete
        # - the text has a 'tick' effect 
        # - there are characters of text left to tick

        if not self.pages:
            return

        self._currentPage = self.pages[self._pageIndex]
        self._textPosition = self._currentPage.textOffset 

        if self.complete is False and self.textEffect == 'tick' and self._characterIndex < len(self.pages[self._pageIndex].text):
            self._tickTimer += deltaTime
            # advance the tick effect
            if self._tickTimer >= self.tickSpeed:
                self._tickTimer = 0
                self._characterIndex += 1
                # play a sound if one has been specified
                if self.tickSound is not None:
                    pygame.mixer.Sound.play(self.tickSound)

    def draw(self, surface):

        '''
        Draws the dialogue box, along with the current page of text.

        :param float deltaTime: the game time elapsed since the last frame (default = 1).
        '''

        # only draw if visible
        if self.visible is False:
            return
    
        # only draw if pages exist
        if not self.pages:
            return
        
        # use custom draw method instead, if one if specified
        if self.customDrawMethod is not None:
            self.customDrawMethod(surface)
            return

        # clear the main surface
        self._surface.fill((0, 0, 0, 0), (0, 0, *self.size))

        # draw background
        if self.backgroundColor is not None:
            pygame.draw.rect(
                self._surface,
                self.backgroundColor,
                (0, 0, *self.size),
                border_radius=self.borderRadius
            )
        
        # draw border
        if self.borderWidth > 0 and self.borderColor is not None:
            pygame.draw.rect(
                self._surface,
                self.borderColor,
                (0, 0, *self.size),
                width = self.borderWidth,
                border_radius=self.borderRadius
            )   

        #
        # draw texture and border, etc.
        #

        # if a texture exists for the current dialogue page
        if self._currentPage.texture is not None:

            # draw a background for the texture if one is specified
            if self._currentPage.textureBackgroundColor is not None:

                # draw texture background
                pygame.draw.rect(self._surface, self._currentPage.textureBackgroundColor,
                                 (*self._currentPage.texturePosition,
                                  self.size[1] - 2 * self.borderWidth - 2 * self.padding,
                                  self.size[1] - 2 * self.borderWidth - 2 * self.padding),
                                  border_radius=self.borderRadius
                )

                # draw texture
                self._surface.blit(
                    pygame.transform.scale(self._currentPage.texture, self._currentPage.textureSize),
                    (self._currentPage.texturePosition[0] + self.borderWidth + self.padding + self._currentPage.texturePadding[0],
                     2 * self.borderWidth + 2 * self.padding + self._currentPage.texturePadding[1])
                )

                # draw texture border
                if self.borderWidth > 0 and self.borderColor is not None:
                    pygame.draw.rect(self._surface, self.borderColor,
                                    (*self._currentPage.texturePosition,
                                    self.size[1] - 2 * self.borderWidth - 2 * self.padding,
                                    self.size[1] - 2 * self.borderWidth - 2 * self.padding),
                                    width = self.borderWidth,
                                    border_radius=self.borderRadius
                    )

        #
        # draw each split line of text
        #

        if self.textEffect == 'none':

            heightOffset = 0
            for line in self._currentPage.splitText:

                drawText(
                    self._surface,
                    line,
                    (self._textPosition[0], self._textPosition[1] + heightOffset),
                    self.font,
                    color = self.textColor
                )

                heightOffset += self.font.get_height() + self.padding // 2
        
        # if 'tick', only draw currently visible characters
        if self.textEffect == 'tick':

            charsLeft = self._characterIndex

            heightOffset = 0
            for line in self._currentPage.splitText:

                if charsLeft > 0:

                    drawText(
                        self._surface,
                        line[0:charsLeft],
                        (self._textPosition[0], self._textPosition[1] + heightOffset),
                        self.font,
                        color = self.textColor
                    )

                    charsLeft -= len(line)

                heightOffset += self.font.get_height() + self.padding // 2 

        # draw the dialogue surface to the surface passed
        surface.blit(self._surface, (self.position[0], self.position[1]))     
    
    def addPage(
        self,
        text,
        texture = None,
        textureAlignment = 'left',
        textureBackgroundColor = None
    ):

        '''
        Adds a page of dialogue to the dialogue box.

        :param str text: The text to display.
        :param pygame.texture texture: The texture to display with the text (default = None).
        :param str textureAlignment: Position of the texture ('left' or 'right', default = 'left').
        :param pygame.Color textureBackgroundColor: The texture background color (default = None).
        '''

        # create and add a new DialoguePage, setting this object as the 'parent'
        self.pages.append(
            DialoguePage(
                self,
                text,
                texture,
                textureAlignment,
                textureBackgroundColor
            )
        )

        # add the text advance symbol to all but the last pages,
        # and remove from the final page
        if self.advanceTextSymbol is not None:

            # add the symbol to all but the last page, if it hasn't been added already
            for i in range(0, len(self.pages) - 1):
                if self.pages[i].text[-len(self.advanceTextSymbol):] != self.advanceTextSymbol:
                    self.pages[i].text = self.pages[i].text + ' ' + self.advanceTextSymbol

            # remove the symbol from the final page, but only if it needs adding
            if self.pages[-1].text[-len(self.advanceTextSymbol):] == self.advanceTextSymbol:
                self.pages[-1].text = self.pages[-1].text[:-len(self.advanceTextSymbol)-1]
        
    def advance(self):
    
        '''
        Advances the dialogue. This means firstly showing all text for the current page, and then
        moving to the next page of dialogue.
        '''

        if not self.pages:
            return

        self._tickTimer = 0

        # if there is still more text to display on the current page
        # then advance to the end of that page's text
        if self.textEffect == 'tick' and self._characterIndex < len(self._currentPage.text):
            self._characterIndex = len(self._currentPage.text)
        # otherwise, move to the next page of text and move to the first character
        else:
            self._pageIndex += 1
            self._characterIndex = 0
            
            # if at the last page then hide the box and mark as completed
            if self._pageIndex > len(self.pages) - 1:
                self._pageIndex = len(self.pages) - 1
                self.visible = False
                self.complete = True

    def getText(self, split = True):

        '''
        Get the current page of text. If the text has a 'tick' effect,
        only the currently visible text is returned.

        :param bool split: Returns a list of text, split according to the available
        text width (default = True).
        '''

        if split == True:

            if self.textEffect == 'tick':

                charsLeft = self._characterIndex
                textLines = []

                for line in self._currentPage.splitText:
                    if charsLeft > 0:
                        textLines.append(line[0:charsLeft])
                        charsLeft -= len(line)
                
                return textLines

            else:
                return self._currentPage.splitText

        else:

            if self.textEffect == 'tick':
                return self._currentPage.text[:self._characterIndex]
            else:
                return self._currentPage.text            




#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

def splitTexture(texture, newTextureSize, rows = 0, columns = 0, clippingRect = None):

    '''
    Split a texture into a 2D list of sub-textures, using the given size.
    (Use pygamepal.flatten() to flatten the list returned into a single list.)

    :param pygame.texture texture: The original texture to split.
    :param (int, int) newTextureSize: The size (w, h) of each of the new textures.
    :param int rows: The number of rows down to split (default = 0, which means split all available rows).
    :param int columns: The number of columns scross to split (default = 0, which means split all available columns).
    :param pygame.Rect clippingRect: The (x, y, w, h) portion of the source image to split (default = None, which uses all of the source image).
    :return list(list(pygame.texture)) newTextures: The 2D list of new textures.
    '''

    # clip the texture if a clipping rectangle is specified
    if clippingRect is not None:
        texture = texture.subsurface(
            max(0, clippingRect[0]),
            max(0, clippingRect[1]),
            min(clippingRect[2], texture.get_width()),
            min(clippingRect[3], texture.get_height())
        )

    # clip the texture if a number of rows and/or columns are specified
    r = texture.get_height()
    if rows > 0:
        r = min(rows * newTextureSize[1], texture.get_height())
    c = texture.get_width()
    if columns > 0:
        c = min(columns * newTextureSize[0], texture.get_width())
    texture = texture.subsurface((0, 0, c, r))

    # the list of textures to return
    newTextures = []

    # use texture.subsurface to loop through the image
    for row in range(0, texture.get_height(), newTextureSize[1]):

        if row + newTextureSize[1] <= texture.get_height():

            newRow = []
            for column in range(0, texture.get_width(), newTextureSize[0]):
                
                if column + newTextureSize[0] <= texture.get_width():

                    # add the cropped texture to the list
                    newRow.append(texture.subsurface(
                        column, row, newTextureSize[0], newTextureSize[1]))
                
            newTextures.append(newRow)
            
    return newTextures


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class Particle():

    '''
    A single particle created by a particle emitter.
    (There is no need to use this class directly, as particle emitters create particles themselves.)
    
    :param (int, int) position: The top-left (x, y) position of the particle.
    :param (float, float) velocity: The (x, y) velocity vector, with (0, 0) being stationary.
    :param (float, float) acceleration: The (x, y) acceleration vector, with (0, 0) = no acceleration.
    :param int lifetime: The time after which the particle is deleted.
    :param int size: The radius of the particle.
    :param float sizeDecay: The amount to decrement the size each frame.
    :param pygame.Color color: The particle color.
    '''

    def __init__(self, position, velocity, acceleration, lifetime, size, sizeDecay, color):
        self.position = position
        self.velocity = velocity
        self.acceleration = acceleration
        self.lifetime = lifetime
        self.size = size
        self.sizeDecay = sizeDecay
        self.color = color

        self.finished = False

    def update(self, deltaTime=1):

        '''
        Update method called once per frame by the owner particle emitter.

        :param float deltaTime: The time since the last update (default = 1).
        '''
        
        self.size -= self.sizeDecay * deltaTime
        self.lifetime -= deltaTime
        # update velocity with respect to acceleration
        self.velocity = (
            (self.velocity[0] + self.acceleration[0] * deltaTime),
            (self.velocity[1] + self.acceleration[1] * deltaTime) 
        )
        # update position with respect to velocity
        self.position = (
            self.position[0] + self.velocity[0] * deltaTime,
            self.position[1] + self.velocity[1] * deltaTime 
        )

    def draw(self, surface):

        '''
        Draw method called by the owner particle emitter.

        :param pygame.Surface surface: The surface to draw to.
        '''
        
        pygame.draw.circle(
            surface,
            self.color,
            self.position,
            self.size)


#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

from .drawText import splitText

class DialoguePage:

    '''
    Pages for the Dialogue class
    (Users of the PygamePal library should not need to use this class directly. Instead, it is used by pygamepal.Dialogue).

    :param pygamepal.Dialogue parent: The Dialogue object in which the DialoguePage exists.
    :param str text: The page of text.
    :param pygame.Texture texture: The texture to display with the text (default = None).
    :param string alignment: The alignment of the texture (either 'left' or 'right', default = 'left').
    :param pygamepal.Color textureBackgroundColor: The color to disply behind the texture (default = None).
    '''

    def __init__(
        self,
        parent,
        text,
        texture = None,
        textureAlignment = 'left',
        textureBackgroundColor = None
    ):
        
        self.parent = parent
        self.texture = texture
        self.textureAlignment = textureAlignment
        self.textureBackgroundColor = textureBackgroundColor
        self.text = text

        self.calculateTextOffset()
        self.calculateTextureSize()
        self.calculateTexturePosition()
    
    def calculateTextOffset(self):
        self.textOffset = (self.parent.borderWidth + self.parent.padding, self.parent.borderWidth + self.parent.padding)
        if self.texture is not None and self.textureAlignment == 'left':
            self.textOffset = (
                self.textOffset[0] + (self.parent.size[1] - 2*self.parent.borderWidth - self.parent.padding),
                self.textOffset[1])
    
    def calculateTextureSize(self):
        
        '''
        Calculates the (w, h) size of the texture, based on the available space.
        '''
        
        if self.texture is None:
            return
        
        textureSpaceAvailable = self.parent.size[1] - 4 * self.parent.padding - 4 * self.parent.borderWidth
        _, _, w, h = self.texture.get_rect()
        factor = textureSpaceAvailable / max(w, h)
        self.textureSize = (w * factor, h * factor)
        self.texturePadding = (
            (textureSpaceAvailable - self.textureSize[0]) / 2,
            (textureSpaceAvailable - self.textureSize[1])
        )
    
    def calculateTexturePosition(self):

        '''
        Calculates the (x, y) position of the texture.
        '''

        if self.texture is None:
            return
        
        # calculate position from left
        leftPos = self.parent.borderWidth + self.parent.padding
        if self.textureAlignment == 'right':
            leftPos = self.parent.size[0] - self.parent.size[1] + self.parent.padding + self.parent.borderWidth

        # calculate position from top
        topPos = self.parent.borderWidth + self.parent.padding

        self.texturePosition = (leftPos, topPos)

    def calculateTextWidth(self):

        '''
        Calculates the available width for the text.
        '''

        # width available = parent Dialogue box size - parent Dialogue borders and padding
        self.textWidthAvailable = self.parent.size[0] - 2 * self.parent.borderWidth - 2 * self.parent.padding

        # subtract texture width (if texture exists), including borders and padding
        if self.texture is not None:
            self.textWidthAvailable -= (self.parent.size[1]) - (2 * self.parent.padding) - (2 * self.parent.borderWidth)

    @property
    def text(self):
        '''
        Get / set the page text.
        '''
        return self._text
    
    @text.setter
    def text(self, value):
        self._text = value
        self.calculateTextWidth()
        self.splitText = splitText(self.text, font = self.parent.font, width = self.textWidthAvailable)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import random
from .particle import Particle

class ParticleEmitter():

    '''
    A marticle emitter creates particles for the duration of its lifetime, with the properties specified.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/particlesExample.gif?raw=true
    
    `Example Particles code`_.

    .. _Example Particles code: https://github.com/rik-cross/pygamepal/blob/main/examples/particlesExample.py
    
    Particle emitter attributes:

    :param (int, int) emitterPosition: The (x, y) top-left position of the emitter (default = (0, 0)).
    :param (int, int) emitterSize: The (w, h) size of the emitter. Particles are created randomly within the emitter bounds (default = (0, 0)).
    :param int emitterLifetime: The emitter only emits particles during its lifetime (-1 = forever, default = 100).
    :param (float, float) emitterVelocity: The (x, y) velocity of the emitter (default = no velocity (0, 0)).
    :param (float, float) emitterAcceleration: The (x, y) acceleration of the emitter (default = no acceleration (0, 0)).
    :param float emitterParticleDelay: The delay between particles emitted (default = 5).

    Particle attributes:

    :param (float, float) particleVelocityMin: The minimum (x, y) velocity of a particle (default = (-1, -1)). Emitted particles are given a random velocity between the minimum and maximum.
    :param (float, float) particleVelocityMax: The maximum (x, y) velocity of a particle (default = (1, 1)). Emitted particles are given a random velocity between the minimum and maximum.
    :param (float, float) particleAccelerationMin: The minimum (x, y) acceleration of a particle (default = (0, 0)). Emitted particles are given a random acceleration between the minimum and maximum.
    :param (float, float) particleAccelerationMax: The maximum (x, y) acceleration of a particle (default = (0, 0)). Emitted particles are given a random acceleration between the minimum and maximum.
    :param int particleLifetime: Each particle is destroyed at the end of its lifetime (default = 100).
    :param int particleSize: The radius of each particle (default = 20).
    :param float particleSizeDecay: The amount to decrement the particle size each frame / per second (default = 0.2).
    :param list(pygame.Color) particleColors: Emitted particles are randomly given a color from the list (default = ['white']).
    '''

    def __init__(self,
        # emitter attributes
        emitterPosition = (0, 0),
        emitterSize = (0, 0),
        emitterLifetime = 100,
        emitterVelocity = (0, 0),
        emitterAcceleration = (0, 0),
        emitterParticleDelay = 5,
        # particle attributes
        particleVelocityMin = (-1, -1),
        particleVelocityMax = (1, 1),
        particleAccelerationMin = (0, 0),
        particleAccelerationMax = (0, 0),
        particleLifetime = 100,
        particleSize = 20,
        particleSizeDecay = 0.2,
        particleColors = ['white']
    ):
        
        self.particleList = []
        
        self.position = emitterPosition
        self.size = emitterSize
        self.lifetime = emitterLifetime
        self.acceleration = emitterAcceleration
        self.velocity = emitterVelocity
        
        self.particleVelocityMin = particleVelocityMin
        self.particleVelocityMax = particleVelocityMax
        self.particleAccelerationMin = particleAccelerationMin
        self.particleAccelerationMax = particleAccelerationMax
        self.particleDelay = emitterParticleDelay

        self.particleLifetime = particleLifetime
        self.particleSize = particleSize
        self.sizeDecay = particleSizeDecay
        self.timeSinceLastParticle = 0
        self.particleColors = particleColors

        # this is set to true once the emitter lifetime is reached
        self.finished = False
    
    def update(self, deltaTime = 1):
        
        '''
        Must be called once per frame to update the particle emitter and its particles.

        :param float deltaTime: The time elapsed since the last update (default = 1).
        '''

        if self.finished:
            return
        
        #
        # update this particle emitter
        #

        if self.lifetime > 0:
            self.lifetime -= deltaTime
        
        # transform
        self.velocity = (self.velocity[0] + self.acceleration[0] * deltaTime,
                         self.velocity[1] + self.acceleration[1] * deltaTime)
        
        self.position = (self.position[0] + self.velocity[0] * deltaTime,
                         self.position[1] + self.velocity[1] * deltaTime)

        #
        # emit more particles
        #

        self.timeSinceLastParticle += 1

        if self.lifetime > -1 and self.lifetime <= 0 and len(self.particleList) == 0:
            self.finished = True

        # create new particles
        if self.timeSinceLastParticle >= self.particleDelay and (self.lifetime == -1 or self.lifetime > 0):

            self.timeSinceLastParticle = 0

            # choose random acceleration, velocity and position between limits
            randomAcceleration = (random.uniform(self.particleAccelerationMin[0], self.particleAccelerationMax[0]),
                                  random.uniform(self.particleAccelerationMin[1], self.particleAccelerationMax[1]))
            randomVelocity = (random.uniform(self.particleVelocityMin[0], self.particleVelocityMax[0]),
                              random.uniform(self.particleVelocityMin[1], self.particleVelocityMax[1]))
            # particle can emit from anywhere within the emitter bounds
            randomPosition = (random.uniform(self.position[0], self.position[0] + self.size[0]),
                              random.uniform(self.position[1], self.position[1] + self.size[1]))
            
            # add new particle to the list
            self.particleList.append(Particle(
                lifetime = self.particleLifetime,
                acceleration = randomAcceleration,
                position = randomPosition,
                velocity = randomVelocity,
                color = random.choice(self.particleColors),
                size = self.particleSize,
                sizeDecay = self.sizeDecay
            ))

        # update each particle
        for p in self.particleList:
            p.update(deltaTime)
            if p.size <= 0 or p.lifetime <= 0:
                self.particleList.remove(p)

    def draw(self, surface):

        '''
        Must be called once per frame to draw the emitter's particles.

        :param pygame.Surface surface: The surface to draw to.
        '''

        if self.finished:
            return

        for p in self.particleList:
            p.draw(surface)